/* rtsyscall.c   
 * "light weight" system calls
 */


#include <asm/errno.h>
#include <asm/rtcontext.h>
#include <asm/rtthread.h>
#include <asm/segment.h>
#include <linux/rtsched.h>
#include <linux/rtfifo.h>
#include <asm/rttime.h>
#include <asm/io.h>


#include <linux/cons.h>

/* misc. syscalls  */

asmlinkage void sys_RTmem_write(int val, int addr) {
 *((int *)addr) = val;
}

asmlinkage int sys_RTmem_read(int val, int addr) {
 return *((int *)addr);
}

asmlinkage void sys_RTport_writeb(int val, int addr) {
 outb(val, addr);
}

asmlinkage void sys_RTport_writew(int val, int addr) {
 outw(val, addr);
}

asmlinkage int sys_RTport_readb(int addr) {
 return inb(addr);
}

asmlinkage int sys_RTport_readw(int addr) {
 return inw(addr);
}

asmlinkage int sys_RTget_time( RTime *time) {
 RTime t=RTget_time();
 memcpy_tofs( time, &t, sizeof( RTime));
 return 0;
}


int RTsys_puts(char *s) {
 static char buf[80];
 memcpy_fromfs( buf, s, 79);
 buf[79]=0;
 conpr(buf);
 return 0;
}



int RTsys_nothing( void) {
 return -ENOSYS;
}


__asm__ (".text\n\t"
	__ALIGN_STR "\n\t"
	".globl "SYMBOL_NAME_STR(RTsystem_call)", "
            SYMBOL_NAME_STR(RTret_from_sys_call)"\n\t"
SYMBOL_NAME_STR(RTsystem_call)":\n\t"
	SAVE_RT_CONTEXT
	"cli\n\t"
	"cmpl "SYMBOL_NAME_STR(RTno_of_sys_calls)", %eax\n\t"
	"jb 2f\n\t"
	"movl $-" STR( ENOSYS) "," STR( RT_EAX) "(%esp)\n\t"
	"jmp "SYMBOL_NAME_STR(RTret_from_sys_call)"\n"

"2:	call "SYMBOL_NAME_STR(RTsys_call_table)"(,%eax,4)\n\t"
	"movl %eax," STR(RT_EAX) "(%esp)\n\t"
	"jmp "SYMBOL_NAME_STR(RTret_from_sys_call)"\n\t"

	__ALIGN_STR "\n\t"
SYMBOL_NAME_STR(RTret_from_sys_call)":\n\t"
	"cli\n\t"
	"mov "SYMBOL_NAME_STR(RTcurrent)", %ecx\n\t"
	"mov "SYMBOL_NAME_STR(RTnew_task)", %edx\n\t"
	"cmpl %ecx, %edx\n\t"
	"je 1f\n\t"

	"mov SFIF, %eax\n\t"
	"mov %eax, 0x24( %ecx)\n\t"

	"mov %cr3, %eax\n\t"
	"mov %eax, 0x1C( %ecx)\n\t"
	"mov %ss , 0x08( %ecx)\n\t"
	"mov %esp, 0x04( %ecx)\n\t"
	"mov %edx, "SYMBOL_NAME_STR(RTcurrent)"\n\t"
	"mov 0x08( %edx), %ss \n\t"
	"mov 0x04( %edx), %esp\n\t"
	"mov 0x1C( %edx), %eax\n\t"
	"mov %eax, %cr3\n\t"

	"mov 0x24( %edx), %eax\n\t"
	"mov %eax, SFIF\n\t"

"1:"	RESTORE_RT_CONTEXT
	"iret\n\t"

);


void RTinit_thread( struct RTthread *tss, long code, 
	long kernelstack, long userstack, struct RTmm_info *mm) {

 long * kst=( long *)kernelstack;


 *--kst=RT_USER_DS;
 *--kst = userstack;    /* not used anymore */
 *--kst=0x0200;		/* eflags:  IF=1; IOPL=0 */
 *--kst=RT_USER_CS;
 *--kst=code;

 *--kst=RT_USER_DS;	/* fs */
 *--kst=RT_USER_DS;	/* es */ 
 *--kst=RT_USER_DS;	/* ds */ 
 *--kst=0;		/* eax */ 
 *--kst=0;		/* ebp */ 
 *--kst=0;		/* edi */ 
 *--kst=0;		/* esi */ 
 *--kst=0;		/* edx */ 
 *--kst=0;		/* ecx */ 
 *--kst=0;		/* ebx */ 

 tss->esp0 = (long) kst;
 tss->ss0 = KERNEL_DS; 
 tss->cr3 = mm->pgdir;
 tss->eflags = 0;  /* Linux interrupts are disabled */
}


void * RTsys_call_table[] = {
 RTsys_nothing,
 RTsys_nothing,
 RTsys_puts,
 RTsys_getpid,
 RTsys_run,
 RTsys_wait_period,
 RTsys_set_params,
 RTsys_wait_start,
 RTsys_kill,
 RTsys_fifo_put,
 RTsys_fifo_get,
 RTsys_mmap
};

int RTno_of_sys_calls = sizeof( RTsys_call_table)/4;

