#include <asm/rtmm.h>
#include <asm/errno.h>
#include <asm/segment.h>
#include <linux/string.h>
#include <linux/rtsched.h>

#define PG_SIZE  0x1000
#define PAGE_TABLE ( PG_SIZE * 1024)

#define USER_LIMIT 0xC0000000

#define PG_MASK (~(PG_SIZE - 1))
#define ADDR_MASK 0xfffff000


inline void set_entry( unsigned long dir, int entry, unsigned long value) {
 ((unsigned long *) dir) [entry] = value;
}


inline unsigned long get_entry( unsigned long dir, int entry) {
 return ((unsigned long *) dir) [entry];
}

#define dir_entry(x)  ( (x >> 22) & 0x3ff)
#define tab_entry(x)  ( (x >> 12) & 0x3ff)

 
extern unsigned long swapper_pg_dir[1024];


inline void zero_page( void * addr) {
 memset( addr, 0, PG_SIZE);
}



int RTmm_init( struct RTmm_info * mm) {
 unsigned long pgdir;

 pgdir = RTget_page();
 if ( !pgdir) {
   return -ENOMEM;
 }
 
 zero_page( (long *) pgdir);

 memcpy( (long *) pgdir + 768, swapper_pg_dir + 768, 256 * 4);

 memcpy( (long *) pgdir, swapper_pg_dir + 768, 256 * 4);  /* ugly, but for now ... */

 mm->pgdir = pgdir;
 return 0;
}



/* if to >= 0x40000000 ( 1G), allocate a new page */

inline int RTmm_map_page( struct RTmm_info * mm, unsigned long to, unsigned long from) {
  int pg_table = get_entry( mm->pgdir, dir_entry( from) ) & ADDR_MASK;
 
  if (!pg_table) {
      pg_table = RTget_page();
      if (!pg_table) return -ENOMEM;
      zero_page( ( void *) pg_table);
      set_entry( mm->pgdir, dir_entry( from), pg_table + 7);
  }

  if ( to >= 0x40000000) {
      to = RTget_page();
      if (!to) return -ENOMEM;
      zero_page( ( void *) to);
      set_entry( pg_table, tab_entry(from), to + 7);
  }
  else  if ( (get_entry( pg_table, tab_entry(from)) & ADDR_MASK) != to) {
      return -EINVAL;
  }
  return 0;
}


int RTmm_map_user( struct RTmm_info * mm, unsigned long addr, unsigned long len) {
 unsigned long first_page = addr & PG_MASK;
 unsigned long last_page = (addr+len-1) & PG_MASK;
 unsigned long i;
 unsigned long ret;

 if ( first_page >= USER_LIMIT || last_page >= USER_LIMIT || first_page > last_page) {
   return -ENOMEM;
 }

 for ( i =  first_page; i <= last_page; i++) {
   ret = RTmm_map_page( mm, 0x80000000, i); 
   if (ret) {
     return ret;
   }
 }
 return 0;
}


int RTmm_alloc_user( struct RTmm_info * mm, unsigned long len, 
				unsigned long * result) {

 unsigned long pgtable = 0;

 /* we want to use one page table */
 if ( len > PAGE_TABLE) {
   return -ENOMEM;
 }

 /* there will always be 0 */
 while ( get_entry( mm->pgdir, pgtable)) {
   pgtable++;
 }

 *result = pgtable * PAGE_TABLE;
 return RTmm_map_user( mm, *result, len); 
}


void RTmm_free_user( struct RTmm_info *mm) {
 /* nothing for now */

}



int RTsys_mmap( unsigned long to, unsigned long len, unsigned long *result) {
 unsigned long first_page = to & PG_MASK;
 unsigned long last_page = (to+len-1) & PG_MASK;
 unsigned long res=0;
 unsigned long ret;
 unsigned long i;

 /* we want to use ONE page table */
 if ( len > PAGE_TABLE) {
   return -ENOMEM;
 }

 /* there will always be 0 */
 while ( get_entry( RTcurrent->mm.pgdir, res)) {
   res++;
 }

 res *= PAGE_TABLE;
  
 for ( i = 0; i <= last_page - first_page; i++) {
   ret = RTmm_map_page( &RTcurrent->mm, i + first_page, res + i); 
   if (ret) {
     return ret;
   }
 }

 res += ( to & (~PG_MASK));  /* offset */
 put_user( res, result);
 
 return 0;
}

