/*
 *  arch/i386/kernel/rttime.c  
 *
 * (C) 1996 Michael Barabanov
 *
 */

#include <asm/rttime.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/sched.h>
#include <linux/cons.h>
#include <linux/mc146818rtc.h>

#define RT_TIMER_IRQ 0
#define RT_CLOCK_IRQ 8

#define LATCH  0x10000

#define TIMER_REACH 65536

volatile RTime rtime=0;
volatile int last_count=0;

volatile static int timer_on=0;
static RTime timer_expires;


static inline int read_cnt2( void) {
 int a;
 __asm__ (
	"mov $0x80, %%al\n\t" 
	"outb %%al, $0x43\n\t"
	"inb $0x42, %%al\n\t"
	"mov %%al, %%dl\n\t"
	"inb $0x42, %%al\n\t"
	"mov %%al, %%ah\n\t"
	"movb %%dl,%%al\n\t"
	"andl $0xffff, %%eax\n\t"
	 : "=a" (a): : "dx");
 return a;
}


int RTset_timer_routine( void (* handler)(void)) {
 free_RTirq( RT_TIMER_IRQ);
 return request_RTirq( RT_TIMER_IRQ, handler);
}


inline RTime RTget_time( void) {
 int cnt = read_cnt2();
 return rtime + (( last_count - cnt) & ( LATCH - 1 ));
}


static inline void try_to_set( void) {
 int init_count;
 static int i=0;
/*
 if ( timer_expires != RT_END && i<200) {
  printk("rtime=%x%08x  timer_on=%d\n", (int) (rtime>>32), (int) rtime, timer_on);
  i++;
 }
*/

 if ( !timer_on && rtime + TIMER_REACH > timer_expires) {

   outb( 0xff, 0x40);  /* MSB 8253 was waiting for */
   timer_on=1;
 
   init_count = timer_expires - RTget_time();
   outb( init_count, 0x40);
   outb( init_count >> 8, 0x40);

   timer_expires = RT_END;	/* this will prevent setting this value twice */
 }
}


inline void RTreset_timer( void) {
 if ( timer_on) {
   outb( 0xff, 0x40);		/* 8253 will wait for MSB */
 
   outb_p(0x0a, 0x20);		/* check for pending timer interrupt */

   if (inb(0x20) & 1) {		/* clear pending irq */
     outb( 0x08, 0x20);		/* poll command */
     (void ) inb ( 0x20);	/* we know what's in here ... */
     outb( 0x20 , 0x20);	/* EOI */
   }
   timer_on=0;
 }
 timer_expires = RT_END;
}


void RTset_timer( RTime when) {
 RTreset_timer();   
 timer_expires = when;
 try_to_set();
}


void RTCinterrupt( void) {
 int new_count;

 (void) CMOS_READ( RTC_REG_C);        /* clear IRQ */ 
 
 new_count = read_cnt2();
 rtime += ( last_count - new_count) & ( LATCH - 1);
 last_count = new_count; 

 try_to_set();
}


void RTtime_init( void) {
  char ctemp;

  /* program channel 2 of the 8253 chip for periodic interrrupt */
  outb_p(0xb4,0x43);			/* binary, mode 2, LSB/MSB, ch 2 */
  outb_p(LATCH & 0xff , 0x42);		/* LSB */
  outb( (LATCH >> 8) & 0xff , 0x42);	/* MSB */
  ctemp=inb(0x61);            
  ctemp &= 0xfd;                  /* shut up the speaker */
  ctemp |= 1;                     /* enable counting ( GATE2 = 1; ) */
  outb( ctemp,0x61);


  /* program channel 0 of the 8253 chip as an one-shot timer */
  enable_irq( RT_TIMER_IRQ);
  outb( 0x30 , 0x43);  /* ch 0, mode 0, lsb+msb */
  outb( 0xff, 0x40);
  outb( 0xff, 0x40);

  timer_on=1;
  RTreset_timer();
 
  timer_expires = RT_END;


  /* program RTC */ 
  CMOS_WRITE(0x29, RTC_REG_A);  	/* 32kHz Time Base, 128Hz interrupt frequency */
  ctemp = CMOS_READ( RTC_REG_B);
  ctemp &= 0x8f;                	/* Clear */
  ctemp |= 0x40;                	/* Periodic interrupt enable */
  CMOS_WRITE( ctemp, RTC_REG_B); 

  request_RTirq( RT_CLOCK_IRQ,&RTCinterrupt);
 
  (void) CMOS_READ(RTC_REG_C);

} 


