#ifndef _ASM_IRQ_H
#define _ASM_IRQ_H

/*
 *	linux/include/asm/irq.h
 *
 *	(C) 1992, 1993 Linus Torvalds
 *
 *      RT-version by Michael Barabanov
 *
 *	IRQ/IPI changes taken from work by Thomas Radke <tomsoft@informatik.tu-chemnitz.de>
 */

#include <linux/linkage.h>
#include <asm/segment.h>
#include <asm/rtcontext.h>

#define NR_IRQS 16

extern void disable_irq(unsigned int);
extern void enable_irq(unsigned int);

#define S_CLI \
	"movl $0,"SYMBOL_NAME_STR(SFIF)"\n\t"

#define S_IRET \
        SAVE_LEAST \
	"cli\n\t \
	movl "SYMBOL_NAME_STR(SFREQ)",%edx\n\t \
	andl "SYMBOL_NAME_STR(SFMASK)",%edx\n\t \
	bsrl %edx,%eax\n\t \
	jz 991f\n\t" \
	S_CLI \
	"sti\n\t \
 	jmp  "SYMBOL_NAME_STR(SFIDT)"(,%eax,4)\n\t \
991:	movl $1,"SYMBOL_NAME_STR(SFIF)"\n\t \
	sti\n\t" \
	RESTORE_LEAST \
	"iret\n\t"

#define S_INT \
	RESTORE_LEAST

#define S_STI \
	"sti\n\t\
	pushfl\n\t \
	pushl $" STR(KERNEL_CS) "\n\t \
        pushl $992f\n\t" \
        S_IRET \
"992:\n\t"	

#define SAVE_ALL \
	"cld\n\t" \
	"push %gs\n\t" \
	"push %fs\n\t" \
	"push %es\n\t" \
	"push %ds\n\t" \
	"pushl %eax\n\t" \
	"pushl %ebp\n\t" \
	"pushl %edi\n\t" \
	"pushl %esi\n\t" \
	"pushl %edx\n\t" \
	"pushl %ecx\n\t" \
	"pushl %ebx\n\t" \
	"movl $" STR(KERNEL_DS) ",%edx\n\t" \
	"mov %dx,%ds\n\t" \
	"mov %dx,%es\n\t" \
	"movl $" STR(USER_DS) ",%edx\n\t" \
	"mov %dx,%fs\n\t"   \
	"movl $0,%edx\n\t"  \
	"movl %edx,%db7\n\t"

/*
 * SAVE_MOST/RESTORE_MOST is used for the faster version of IRQ handlers,
 * installed by using the SA_INTERRUPT flag. These kinds of IRQ's don't
 * call the routines that do signal handling etc on return, and can have
 * more relaxed register-saving etc. They are also atomic, and are thus
 * suited for small, fast interrupts like the serial lines or the harddisk
 * drivers, which don't actually need signal handling etc.
 *
 * Also note that we actually save only those registers that are used in
 * C subroutines (%eax, %edx and %ecx), so if you do something weird,
 * you're on your own. The only segments that are saved (not counting the
 * automatic stack and code segment handling) are %ds and %es, and they
 * point to kernel space. No messing around with %fs here.
 */
#define SAVE_MOST \
	"cld\n\t" \
	"push %es\n\t" \
	"push %ds\n\t" \
	"pushl %eax\n\t" \
	"pushl %edx\n\t" \
	"pushl %ecx\n\t" \
	"movl $" STR(KERNEL_DS) ",%edx\n\t" \
	"mov %dx,%ds\n\t" \
	"mov %dx,%es\n\t"

#define RESTORE_MOST \
	"popl %ecx\n\t" \
	"popl %edx\n\t" \
	"popl %eax\n\t" \
	"pop %ds\n\t" \
	"pop %es\n\t" \
        S_IRET	

/*
 * The "inb" instructions are not needed, but seem to change the timings
 * a bit - without them it seems that the harddisk driver won't work on
 * all hardware. Arghh.
 */
#define ACK_FIRST(mask) \
	"inb $0x21,%al\n\t" \
	"jmp 1f\n" \
	"1:\tjmp 1f\n" \
	"1:\torb $" #mask ","SYMBOL_NAME_STR(cache_21)"\n\t" \
	"movb "SYMBOL_NAME_STR(cache_21)",%al\n\t" \
	"outb %al,$0x21\n\t" \
	"jmp 1f\n" \
	"1:\tjmp 1f\n" \
	"1:\tmovb $0x20,%al\n\t" \
	"outb %al,$0x20\n\t"

#define ACK_SECOND(mask) \
	"inb $0xA1,%al\n\t" \
	"jmp 1f\n" \
	"1:\tjmp 1f\n" \
	"1:\torb $" #mask ","SYMBOL_NAME_STR(cache_A1)"\n\t" \
	"movb "SYMBOL_NAME_STR(cache_A1)",%al\n\t" \
	"outb %al,$0xA1\n\t" \
	"jmp 1f\n" \
	"1:\tjmp 1f\n" \
	"1:\tmovb $0x20,%al\n\t" \
	"outb %al,$0xA0\n\t" \
	"jmp 1f\n" \
	"1:\tjmp 1f\n" \
	"1:\toutb %al,$0x20\n\t"

#define UNBLK_FIRST(mask) \
	"inb $0x21,%al\n\t" \
	"jmp 1f\n" \
	"1:\tjmp 1f\n" \
	"1:\tandb $~(" #mask "),"SYMBOL_NAME_STR(cache_21)"\n\t" \
	"movb "SYMBOL_NAME_STR(cache_21)",%al\n\t" \
	"outb %al,$0x21\n\t"

#define UNBLK_SECOND(mask) \
	"inb $0xA1,%al\n\t" \
	"jmp 1f\n" \
	"1:\tjmp 1f\n" \
	"1:\tandb $~(" #mask "),"SYMBOL_NAME_STR(cache_A1)"\n\t" \
	"movb "SYMBOL_NAME_STR(cache_A1)",%al\n\t" \
	"outb %al,$0xA1\n\t"

#define IRQ_NAME2(nr) nr##_interrupt(void)
#define LINUX_IRQ_NAME(nr) IRQ_NAME2(linux_IRQ##nr)
#define RTWLINUX_IRQ_NAME(nr) IRQ_NAME2(rtwlinux_IRQ##nr)
#define RT_IRQ_NAME(nr) IRQ_NAME2(rt_IRQ##nr)
#define IRQ_NAME(nr) IRQ_NAME2(IRQ##nr)
#define FAST_IRQ_NAME(nr) IRQ_NAME2(fast_IRQ##nr)
#define BAD_IRQ_NAME(nr) IRQ_NAME2(bad_IRQ##nr)

#ifdef	CONFIG_SMP
#error "SMP is not supported by RT-Linux yet"
#else

/* mud*/	

#define EOI_FIRST \
	"movb $0x20,%al\n\t \
	outb %al,$0x20\n\t" \

#define EOI_SECOND \
	"movb $0x20,%al\n\t \
	outb %al,$0x20\n\t \
	outb %al,$0xA0\n\t" \

#define MASK(i) "(1 <<" #i ")"


/* A SET bit in a SFMASK  ENABLES the corresponding interrupt ( NOT as in 8259A ) */
#define DISABLE(i) "andl $~" MASK(i) ","SYMBOL_NAME_STR(SFMASK)"\n\t"
#define ENABLE(i) "orl $" MASK(i) ","SYMBOL_NAME_STR(SFMASK)"\n\t"

#define BUILD_IRQ(chip,nr,mask) \
asmlinkage void LINUX_IRQ_NAME(nr); \
asmlinkage void RTWLINUX_IRQ_NAME(nr); \
asmlinkage void RT_IRQ_NAME(nr); \
asmlinkage void IRQ_NAME(nr); \
asmlinkage void FAST_IRQ_NAME(nr); \
asmlinkage void BAD_IRQ_NAME(nr); \
__asm__( \
"\n"__ALIGN_STR"\n" \
SYMBOL_NAME_STR(rt_IRQ) #nr "_interrupt:\n\t" \
	SAVE_RT_CONTEXT \
	EOI_##chip \
	"call *("SYMBOL_NAME_STR(RTaction)" + (" #nr " * 4))\n\t" \
	"jmp "SYMBOL_NAME_STR(RTret_from_sys_call)"\n\t" \
"\n"__ALIGN_STR"\n" \
SYMBOL_NAME_STR(rtwlinux_IRQ) #nr "_interrupt:\n\t" \
	SAVE_LEAST \
	EOI_##chip \
	"orl $" MASK(nr) ","SYMBOL_NAME_STR(SFREQ)"\n\t" \
	"call *("SYMBOL_NAME_STR(RTaction)" + (" #nr " * 4))\n\t" \
	"cmpl $1,"SYMBOL_NAME_STR(SFIF)"\n\t" \
	"je 1f\n\t" \
	RESTORE_LEAST \
	"iret\n\t" \
"1:	sti\n\t" \
	RESTORE_LEAST \
	S_IRET \
"\n"__ALIGN_STR"\n" \
SYMBOL_NAME_STR(linux_IRQ) #nr "_interrupt:\n\t" \
	SAVE_LEAST \
	ACK_##chip(mask) \
	"orl $" MASK(nr) ","SYMBOL_NAME_STR(SFREQ)"\n\t" \
	"cmpl $1,"SYMBOL_NAME_STR(SFIF)"\n\t" \
	"je 1f\n\t" \
	RESTORE_LEAST \
	"iret\n\t" \
"1:	sti\n\t" \
	RESTORE_LEAST \
	S_IRET \
"\n"__ALIGN_STR"\n" \
SYMBOL_NAME_STR(IRQ) #nr "_interrupt:\n\t" \
	S_INT \
	"pushl $-"#nr"-2\n\t" \
	SAVE_ALL \
	DISABLE(nr) \
	"andl $~ " MASK(nr) ","SYMBOL_NAME_STR(SFREQ)"\n\t" \
	"incl "SYMBOL_NAME_STR(intr_count)"\n\t"\
	S_STI \
	"movl %esp,%ebx\n\t" \
	"pushl %ebx\n\t" \
	"pushl $" #nr "\n\t" \
	"call "SYMBOL_NAME_STR(do_IRQ)"\n\t" \
	"addl $8,%esp\n\t" \
	S_CLI \
 	"cli\n\t" \
	UNBLK_##chip(mask) \
	ENABLE(nr) \
	"sti\n\t" \
	"decl "SYMBOL_NAME_STR(intr_count)"\n\t" \
	"jmp ret_from_sys_call\n" \
"\n"__ALIGN_STR"\n" \
SYMBOL_NAME_STR(fast_IRQ) #nr "_interrupt:\n\t" \
	S_INT \
	SAVE_MOST \
	DISABLE(nr) \
	"andl $~ " MASK(nr) ","SYMBOL_NAME_STR(SFREQ)"\n\t" \
	"incl "SYMBOL_NAME_STR(intr_count)"\n\t" \
	"pushl $" #nr "\n\t" \
	"call "SYMBOL_NAME_STR(do_fast_IRQ)"\n\t" \
	"addl $4,%esp\n\t" \
	S_CLI \
 	"cli\n\t" \
	UNBLK_##chip(mask) \
	ENABLE(nr) \
	"sti\n\t" \
	"decl "SYMBOL_NAME_STR(intr_count)"\n\t" \
	RESTORE_MOST \
"\n"__ALIGN_STR"\n" \
SYMBOL_NAME_STR(bad_IRQ) #nr "_interrupt:\n\t" \
	S_INT \
	SAVE_MOST \
	DISABLE(nr) \
	"andl $~ " MASK(nr) ","SYMBOL_NAME_STR(SFREQ)"\n\t" \
	RESTORE_MOST);

#endif
#endif
