#include <linux/rtfifo.h>
#include <asm/segment.h>
#include <linux/malloc.h>
#include <asm/errno.h>


struct RTfifo {
 char *buf;
 char *eob;
 int size;
 char *head;
 char *tail;
};

/* amount of data in a fifo */
static inline int bytes( struct RTfifo * fifo) {
 return  fifo->head  >= fifo->tail ?  
		fifo->head - fifo->tail :
		fifo->size - ( fifo->tail - fifo->head); 
}



int RTfifo_create( struct RTfifo *fifo, char * buf, int sz) {

 fifo->buf = buf;
 fifo->size = sz;
 fifo->eob = buf + sz;
 fifo->head = buf;
 fifo->tail = buf;
 
 return 0; 
}


inline int RTfifo_put( struct RTfifo *fifo, char * buf, int count) {
 int till_end;

 till_end = fifo->eob - fifo-> head;
 if ( till_end > count) {
   memcpy_fromfs( fifo-> head, buf, count);
   fifo->head += count;
 }
 else {
  memcpy_fromfs( fifo-> head, buf, till_end);
  memcpy_fromfs( fifo-> buf, buf + till_end, count - till_end);
  fifo-> head = fifo-> buf + count - till_end;
 }

 return count;
}



inline int RTfifo_get( struct RTfifo *fifo, char * buf, int count) {
 int till_end;

 till_end = fifo->eob - fifo-> tail;
 if ( till_end > count) {
   memcpy_tofs( buf, fifo-> tail, count);
   fifo->tail += count;
 }
 else {
  memcpy_tofs( buf, fifo-> tail, till_end);
  memcpy_tofs( buf + till_end, fifo-> buf, count - till_end);
  fifo-> tail = fifo-> buf + count - till_end;
 }

 return count;
}


/* here goes syscall stuff */

struct RTfifo fifoes[ RT_MAX_FIFO];
static char status[ RT_MAX_FIFO];

static inline struct RTfifo * get_addr( unsigned int fifo) {
 if (fifo < RT_MAX_FIFO && status[ fifo] )
   return &fifoes[fifo];
 return 0;
}


int RTsys_fifo_put( unsigned int f, char * buf, int count) {
 struct RTfifo *fifo= get_addr(f);
 if (!fifo) return -EINVAL;

 if ( fifo->size - bytes( fifo) - 1 < count) {
   return -1;
 }
 return  RTfifo_put( fifo, buf, count);
}


int RTsys_fifo_get( unsigned int f, char * buf, int count) {
 struct RTfifo *fifo= get_addr(f);
 if (!fifo) return -EINVAL;

 if ( bytes( fifo) < count) {
   return -1;
 }
 return  RTfifo_get( fifo, buf, count);
}



int sys_RTfifo_create( unsigned int f, int size) {
 char * buf;
 if ( f >= RT_MAX_FIFO) {
  return -EINVAL;
 }
 
 buf= kmalloc( size, GFP_KERNEL);
 if ( !buf) {
   return -ENOMEM;
 }

 RTfifo_create( &fifoes[f], buf, size);

 status[f] =1;
 return 0;
}


int sys_RTfifo_destroy( unsigned int f) {
 struct RTfifo *fifo= get_addr(f);
 if (!fifo) return -EINVAL;

 status[f]=0;
 
 kfree( fifo-> buf);
 return 0;
}


int sys_RTfifo_put( unsigned int f, char * buf, int count) {
 long flags;
 int ret;
 struct RTfifo *fifo= get_addr(f);
 if (!fifo) return -EINVAL;

 r_save_flags( flags);

 if ( fifo->size - bytes( fifo) - 1 < count) {
   r_restore_flags( flags);
   return -1;
 }
 ret = RTfifo_put( fifo, buf, count);

 r_restore_flags( flags);
 return ret;
}



int sys_RTfifo_get( unsigned int f, char * buf, int count) {
 long flags;
 int ret;
 struct RTfifo *fifo= get_addr(f);
 if (!fifo) return -EINVAL;

 r_save_flags( flags);

 if ( bytes( fifo) < count) {
   r_restore_flags( flags);
   return -1;
 }

 r_restore_flags( flags);
 ret = RTfifo_get( fifo, buf, count);
 return ret;
}


void RTfifo_init( void) {
 int i;
 for (i=0; i< RT_MAX_FIFO; i++) {
   status[i]=0;
 }
}

