
#include <linux/kernel.h>
#include <linux/rtsched.h>
#include <linux/errno.h>
#include <linux/elf.h>
#include <linux/string.h>
#include <asm/rtmm.h>
#include <asm/rtthread.h>
#include <asm/segment.h>
#include <asm/rtcontext.h>
#include <linux/fs.h>


#define RT_USER_STACK_SIZE 8192
#define RT_KERNEL_STACK_SIZE 4096


int RTread_file(struct inode *inode, unsigned long offset,
	char * addr, unsigned long count, int to_kmem)
{
	struct file file;
	int result = -ENOEXEC;

	file.f_mode = 1;
	file.f_flags = 0;
	file.f_count = 1;
	file.f_inode = inode;
	file.f_pos = 0;
	file.f_reada = 0;
	file.f_op = inode->i_op->default_file_ops;
	if (file.f_op->open)
		if (file.f_op->open(inode,&file))
			goto end_readexec;
	if (!file.f_op || !file.f_op->read)
		goto close_readexec;
	if (file.f_op->lseek) {
		if (file.f_op->lseek(inode,&file,offset,0) != offset)
 			goto close_readexec;
	} else
		file.f_pos = offset;
	if (to_kmem) {
		unsigned long old_fs = get_fs();
		set_fs(get_ds());
		result = file.f_op->read(inode, &file, addr, count);
		set_fs(old_fs);
	} else {
		result = file.f_op->read(inode, &file, addr, count);
	}
close_readexec:
	if (file.f_op->release)
		file.f_op->release(inode,&file);
end_readexec:
	return result;
}


int do_RTload( const char *file) {
 struct elfhdr elf_hdr;
 struct elf_phdr * elf_ph;
 static char buf[128];
 struct inode * inode;
 int retval;
 int i,j;
 int pid;
 struct RTtask_struct *newtask;
 struct RTmm_info save_mm;
 unsigned long save_fs;

 
 retval=open_namei( file, 0, 0, &inode, NULL);
 if ( retval) {
   return retval;
 }

 save_fs=get_fs();
 set_fs( RT_USER_DS);
 RTmm_save_active( &save_mm);

 retval = RTread_file( inode, 0, buf, 128, 1);
 if ( retval < 0) {
   goto close_inode;
 }

 elf_hdr = *((struct elfhdr *) buf);
 	
 if (elf_hdr.e_ident[0] != 0x7f ||
      strncmp(&elf_hdr.e_ident[1], "ELF",3) != 0 || 
      elf_hdr.e_type != ET_EXEC || 
      (elf_hdr.e_machine != EM_386 && elf_hdr.e_machine != EM_486)) {
          retval =  -ENOEXEC;
          goto close_inode;
 }

 elf_ph = (struct elf_phdr *) RTkmalloc(elf_hdr.e_phentsize * elf_hdr.e_phnum);
 if (elf_ph == NULL) {
     retval =  -ENOMEM;
     goto close_inode;
 }

 retval = RTread_file(inode, elf_hdr.e_phoff, (char *) elf_ph,
   elf_hdr.e_phentsize * elf_hdr.e_phnum, 1);

 if (retval < 0 ) {
   goto free_ph;
 }


#if 0
 printk("ELF executable   start=%x\n", ( unsigned int ) (elf_hdr.e_entry));
 for ( i = 0; i< elf_hdr.e_phnum; i++) {
  printk("hdr #%d   type=%x  offset=%x  VA=%x  filesz=%x  memsz=%x  flags=%x  align=%x\n",
       i, (unsigned int ) (elf_ph[i].p_type), (unsigned int ) (elf_ph[i].p_offset), 
       (unsigned int ) (elf_ph[i].p_vaddr), (unsigned int ) (elf_ph[i].p_filesz), 
       (unsigned int ) (elf_ph[i].p_memsz), 
       (unsigned int ) (elf_ph[i].p_flags), (unsigned int ) (elf_ph[i].p_align));
 }
#endif

 pid = RTtask_alloc();

 if ( pid < 0)  {
  retval =  -EAGAIN;
  goto free_ph;
 }
 newtask= RTtask_ptr(pid);

 newtask -> kernel_stack = RTget_page();
 if ( ! newtask -> kernel_stack ) {
      retval = -ENOMEM;
      goto free_task;
 }
 retval = RTmm_init( &(newtask -> mm));
 if ( retval < 0) {
   goto free_stack;
 }


 RTmm_restore_active( &(newtask -> mm));

 for ( i = 0; i< elf_hdr.e_phnum; i++) {
   if ( elf_ph[i].p_type != PT_LOAD) {
     retval = -ENOEXEC;
     goto free_memory;
   }

   retval = RTmm_map_user( &(newtask->mm), elf_ph[i].p_vaddr, elf_ph[i].p_memsz);
   if ( retval < 0 ) {
     goto free_memory;
   }

   /* RTmm_map_user zeros the area so we don't have to do it for bss */

   retval = RTread_file( inode, elf_ph[i].p_offset, ( char *) elf_ph[i].p_vaddr, 
      elf_ph[i].p_filesz, 0);


   if ( retval < 0) {
     goto free_memory;
   }


#if 0
   printk("Dump for segment %d\n",i);

   for ( j=0; j <  elf_ph[i].p_memsz; j+=4) {
     printk("%08x ", get_user((unsigned int *)(elf_ph[i].p_vaddr+j)) ); 
     if ( j % 8 == 0) printk("\n");
   }
#endif

 }

 /* now at last we got everything we need to create a new task */

 r_save_flags( i);
 r_cli();

 newtask-> state = RT_TASK_STOPPED;
 newtask-> pid = pid;
 RTinit_thread( &(newtask->tss), elf_hdr.e_entry, 
      (long) ( newtask->kernel_stack) + RT_KERNEL_STACK_SIZE,
      0, &(newtask->mm) );

 r_restore_flags(i);


 retval = pid;
 goto free_ph;



 free_memory:
  RTmm_free_user( &(newtask->mm));

 free_stack:
  RTfree_page( newtask -> kernel_stack);
 
 free_task:
  RTtask_free( pid);

 free_ph:
   RTkfree(elf_ph);

 close_inode:
   iput(inode);

 set_fs( save_fs);
 RTmm_restore_active( &save_mm);

 return retval;
}


asmlinkage int sys_RTload( const char * name) {
 char *tmp;
 int ret;

 if ( !suser()) {
   return -EPERM;
 }

 ret = getname( name, &tmp);
 if (ret) {
   return ret;
 }
 ret=do_RTload( tmp);
 putname( tmp);
 return ret;
}
 
